/*****************************************************************************\
*            mdsp.c - Motorola DSP56301 Board PCI Driver for Linux            *
*                                                                             *
* Copyright (C) 1998  Klaus Knopper                                           *
* see mdsp.h for definitions of registers and flags                           *
* #define DEBUG for more verbose output of module                             *
* #define SLOW if you do not care if your system hangs during poll            *
\*****************************************************************************/

#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#error "MDSP must be built as Module. Please #define MODULE."
#endif

#ifdef CONFIG_PCI
#include <linux/pci.h>
#include <linux/bios32.h>
#else
#error "You must enable PCI support in your Kernel."
#endif

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/utsname.h> /* system_utsname */
#include <linux/ioport.h>  /* check_region(), request_region() */
#include <asm/segment.h>
#include <asm/io.h>

#include "mdsp.h" /* System-Includes, register definitions, makros */

/* Constants */
#ifndef PCI_VENDOR_ID_MOTOROLA
#define PCI_VENDOR_ID_MOTOROLA 0x1057
#endif

#ifndef PCI_DEVICE_ID_DSP56301
#define PCI_DEVICE_ID_DSP56301 0x1801
#endif

#define MDSP_MEM_SIZE 0x00010000  /* DSP56301 has 64K memory-mapped   */

#define MAX_STATUS 4096           /* Max size of /dev/mdspstat buffer */

#define MDSP_INT_LEN 4            /* 32-bit transfers                 */

/* Utility Functions */
#define PRINTK(format,a...) printk("MDSP: "format, ##a)
#ifdef DEBUG
#define DPRINTK(format,a...) printk("MDSP_DEBUG: "format, ##a)
#else
#define DPRINTK(format,a...)
#endif

/* For better readability for non-linux programmers: Indicating Names of */
/* kernel functions (Michael wants it)                                   */
#define kernel_readl readl
#define kernel_inl inl
#define kernel_writel writel
#define kernel_outl outl
#define kernel_verify_area verify_area
#define kernel_var_need_resched need_resched
#define kernel_var_current current
#define kernel_var_jiffies jiffies
#define kernel_schedule schedule
#define kernel_put_user put_user
#define kernel_get_user get_user
#define kernel_memcpy_tofs memcpy_tofs
#define kernel_memcpy_fromfs memcpy_fromfs
#define kernel_pcibios_present pcibios_present
#define kernel_pcibios_find_device pcibios_find_device
#define kernel_pcibios_read_config_word pcibios_read_config_word
#define kernel_pcibios_read_config_dword pcibios_read_config_dword
#define kernel_pcibios_write_config_byte pcibios_write_config_byte
#define kernel_pcibios_write_config_word pcibios_write_config_word
#define kernel_pcibios_write_config_dword pcibios_write_config_dword
#define kernel_pcibios_read_config_byte pcibios_read_config_byte
#define kernel_check_region check_region
#define kernel_request_region request_region
#define kernel_release_region release_region
#define kernel_register_chrdev register_chrdev
#define kernel_select_wait select_wait
#define kernel_unregister_chrdev unregister_chrdev


/* Mapping of physical address into kernel space changed from 2.0 to 2.1 */
#if (LINUX_VERSION_CODE < 0x020100)
#define kernel_ioremap vremap
#define kernel_iounmap vfree
#else
#define kernel_ioremap ioremap
#define kernel_iounmap iounmap
#endif

/* Board parameters */
typedef struct _mdsp_info
{
 unsigned char busnr,devnr,irq,irqtype;
 unsigned short class,ready;
 unsigned int boardtype;
 unsigned int base;        /* Base address MDSP registers                  */
 unsigned int basetype;    /* Memory-mapped for DSP56301                   */
 unsigned int hctr,        /* Control Register Address                     */
              hstr,        /* Status Register Address                      */
              hcvr,        /* Host Control Vector Register                 */
              htxr,        /* Transmission Register Address (Input to DSP) */
              hrxr;        /* Receive Register Address   (Output from DSP) */
 struct wait_queue *wait;
} mdsp_info;

static mdsp_info mdsp_board[MDSP_MAX]; /* structs for all boards found */

/* Static variables */
static int   mdsp_major=MDSP_MAJOR;  /* Major Device Number          */
static char *mdsp_name="mdsp";       /* Name of device               */
static char *status_buf = NULL;      /* /dev/mdspstat info buffer    */
static int   status_len, status_ptr; /* Current buffer params for mdspstat */


static unsigned short mdsp_type[]=     /* Supported DSP board types        */
 { PCI_DEVICE_ID_DSP56301 };           /* Currently only DSP56301, please  */
                                       /* add others                       */

/* Basic register read/write functions. Though the DSP56301 currently      */
/* uses memory-mapped access, other cards might be mapped into io-space,   */
/* so both is supported.                                                   */
static __inline void mdsp_write_reg(unsigned int addr, __u32 val,
                                    unsigned int type)
{
 if(type==PCI_BASE_ADDRESS_SPACE_IO) kernel_outl(val,addr);
 else writel(val,addr);
}

static __inline __u32 mdsp_read_reg(unsigned int addr,unsigned int type)
{
 return (type==PCI_BASE_ADDRESS_SPACE_IO)?kernel_inl(addr)
                                         :readl(addr);
}


/***************************** Status Utilities ******************************\
* State of all DSP56xxx cards may be read from /dev/mdspstat (similar to      *
* /dev/sndstat provided by sound driver).                                     *
\*****************************************************************************/
static int put_status (char *s)
{
 int l=strlen(s);
 if(status_len+l>=MAX_STATUS) return 0;
 memcpy(&status_buf[status_len],s,l);
 status_len+=l;
 return 1;
}

static int put_status_int(unsigned int val,int radix)
{
 int l,v;
 static char hx[] = "0123456789abcdef";
 char        buf[11];
 if(!val) return put_status ("0");
 l=0; buf[10]=0;
 while(val)
  {
   v=val%radix; val=val/radix;
   buf[9-l]=hx[v]; l++;
  }
 if (status_len+l>=MAX_STATUS) return 0;
 memcpy(&status_buf[status_len],&buf[10-l],l);
 status_len+=l;
 return 1;
}

static void init_status (void)
{
 int i;
 char *bl=" ";
 char *cr="\n";
 status_ptr = 0;
 put_status("Kernel: ");
 put_status(system_utsname.sysname);  put_status (bl);
 put_status(system_utsname.nodename); put_status (bl);
 put_status(system_utsname.release);  put_status (bl);
 put_status(system_utsname.version);  put_status (bl);
 put_status(system_utsname.machine);  put_status (cr);
 for(i=0;i<MDSP_MAX;i++)
  {
   if(mdsp_board[i].ready)
    {
     unsigned int base=mdsp_board[i].base;
     unsigned int type=mdsp_board[i].basetype;
     put_status("MDSP ");       put_status_int(i+1,10); put_status(":");
     put_status(" bus 0x");     put_status_int(mdsp_board[i].busnr,16);
     put_status(", device 0x"); put_status_int(mdsp_board[i].devnr,16);
     put_status(", class 0x");  put_status_int(mdsp_board[i].class,16);
     put_status(", address 0x");
     put_status_int(mdsp_board[i].base,16);
     put_status(", irq 0x");    put_status_int(mdsp_board[i].irq,16);
     put_status(", control 0x");
     put_status_int(mdsp_read_reg(base+HCTR,type),16);
     put_status(", status 0x");
     put_status_int(mdsp_read_reg(base+HSTR,type),16);
     put_status(", vector 0x");
     put_status_int(mdsp_read_reg(base+HCVR,type),16);
     put_status(cr);
    }
  }
}

static int read_status (char *buf, int count)
{ /* Return at most 'count' bytes from the status_buf. */
  int l=count,c=status_len-status_ptr;
  if(l>c) l=c;
  if(l<=0) return 0;
  memcpy_tofs(buf,&status_buf[status_ptr],l);
  status_ptr+=l;
  return l;
}

/***************************** Polling ***************************************/

/*****************************************************************************\
* poll() waits for the given flag to be set in the given register.            *
* returns 0 on success. There is no timeout limit, though the polling may be  *
* terminated on reception of any non-blocked signal.                          *
\*****************************************************************************/
static int poll(unsigned int address,__u32 flag,unsigned int type)
{
 for(;;) 
  {
   __u32 status=mdsp_read_reg(address,type);
   unsigned int sig;
   if(status&flag) break;
   /* Did we receive any non-blocked signals? (^C etc.) */
   sig=kernel_var_current->signal&~kernel_var_current->blocked;
   if(sig)
    { /* Handle SIGSTOP etc. */
     DPRINTK("Command aborted by signal 0x%x\n",sig);
     return -ERESTARTSYS;
    }
#ifndef SLOW
   /* Be nice to other processes */
   if(kernel_var_need_resched) kernel_schedule();
#endif
  }
 return 0;
}

/************************** VFS functions ***********************************/
static int read_mdsp(struct inode *inode, struct file *file, char *buffer,
                     int count)
{
 int minor=MINOR(inode->i_rdev),boardnr=minor-1;
 if((minor==0)||(minor>0&&minor<=MDSP_MAX&&mdsp_board[boardnr].ready))
  {
   int i,result;
   if(count<=0) return -EINVAL;
   if((result=kernel_verify_area(VERIFY_WRITE,buffer,count))!=0) return result;
   if(minor==0)
    { /* Write status information for all cards */
     return read_status(buffer,count);
    }
   else
    {
     if(!mdsp_board[boardnr].ready) return -ENODEV;
     /* Count must be a multiple of sizeof(__u32) */
     count &= ~(MDSP_INT_LEN-1);
     DPRINTK("Reading %d 32-bit words from DSP.\n",count/sizeof(int));
     for(i=0; i<count; i+=MDSP_INT_LEN) /* 32-bit values */
      {
       __u32 val;
       /* Wait for status flag, then read. HSTR_HRRQ is set by the DSP       */
       /* automatically after write operations to DTXR.                      */
       result=poll(mdsp_board[boardnr].hstr,HSTR_HRRQ,
                   mdsp_board[boardnr].basetype);
       if(result) return i?i:result;    /* return bytes read so far or error */
       val=mdsp_read_reg(mdsp_board[boardnr].hrxr,
                         mdsp_board[boardnr].basetype);
       /* In case of alignment problems, use                               */
       /* kernel_memcpy_tofs(&buffer[i],&val,sizeof(unsigned int));        */
       /* instead of kernel_put_user(val,(__u32 *)&buffer[i]);             */
       kernel_put_user(val,(__u32 *)&buffer[i]);
      }
     return count;
    }
  }
 return -ENODEV;
}

static int write_mdsp(struct inode *inode, struct file *file,
                      const char *buffer, int count)
{
 int minor=MINOR(inode->i_rdev),boardnr=minor-1;
 if(minor>0&&minor<=MDSP_MAX&&mdsp_board[boardnr].ready)
  {
   int i,result;
   if((result=kernel_verify_area(VERIFY_READ, buffer, count))!=0) return result;
   if(count<sizeof(long)) return -EINVAL;
   /* Count must be a multiple of sizeof(__u32). */
   count &= ~(MDSP_INT_LEN-1);
   DPRINTK("Writing %d 32-bit words to DSP.\n",count/MDSP_INT_LEN);
   for(i=0; i<count; i+=MDSP_INT_LEN) /* 32-bit words */
    {
     __u32 val;
     /* In case of alignment problems, use                               */
     /* kernel_memcpy_fromfs(&val,&buffer[i],sizeof(__u32));             */
     /* instead of val=kernel_get_user((__u32 *)&buffer[i]);             */
     val=kernel_get_user((__u32 *)&buffer[i]);
     /* Wait for status flag in HSTR, then write. HTRQ is set by the DSP */
     /* automatically if there is space in FIFO for another 32-bit word. */
     result=poll(mdsp_board[boardnr].hstr,HSTR_HTRQ,
                 mdsp_board[boardnr].basetype);
     if(result) return i?i:result;  /* return bytes read so far or error */
     mdsp_write_reg(mdsp_board[boardnr].htxr,val,
                    mdsp_board[boardnr].basetype);
    }
   return count;
  }
 return -ENODEV;
}

static int select_mdsp(struct inode *inode, struct file *file, int type,
                       select_table *wait)
{
 int minor=MINOR(inode->i_rdev),boardnr=minor-1;
 if(minor>0&&minor<=MDSP_MAX&&mdsp_board[boardnr].ready)
  {
   int flag;
   for(flag=0;flag<2;flag++)
    { /* try two times, pause in between */
     __u32 status=mdsp_read_reg(mdsp_board[boardnr].hstr,
                                mdsp_board[boardnr].basetype);
     switch(type)
      {
       case SEL_IN:  if(status&HSTR_HRRQ) return 1; break;
       case SEL_OUT: if(status&HSTR_HTRQ) return 1; break;
       case SEL_EX:  if(status&HSTR_TRDY) return 1; break;
       default:      return -EINVAL;
      }
     if(flag==0) kernel_select_wait(&mdsp_board[boardnr].wait,wait);
    }
  }
 else return -ENODEV;
 return 0;
}

static int open_mdsp(struct inode *inode, struct file *file)
{
 int minor=MINOR(inode->i_rdev),boardnr=minor-1;
 if((minor==0)||(minor>0&&minor<=MDSP_MAX&&mdsp_board[boardnr].ready))
  {
   MOD_INC_USE_COUNT; /* Increment in-use counter */
   if(minor==0)
    {
     if(status_buf||(status_buf=(char *)vmalloc(MAX_STATUS))==NULL) return -EIO;
     status_len=status_ptr=0;
     init_status();
    }
   DPRINTK("Minor %d opened.\n",minor);
   return 0;
  }
 return -ENODEV;
}

static void release_mdsp(struct inode *inode, struct file *file)
{
 int minor=MINOR(inode->i_rdev),boardnr=minor-1;
 if((minor==0)||(minor>0&&minor<=MDSP_MAX&&mdsp_board[boardnr].ready))
  {
   if(minor==0 && status_buf) {vfree(status_buf); status_buf=NULL; }
   MOD_DEC_USE_COUNT; /* Increment in-use counter */
   DPRINTK("Minor %d closed.\n",minor);
  }
}

static int ioctl_mdsp(struct inode *inode, struct file *file, unsigned int cmd,
                      unsigned long param)
{
 int minor=MINOR(inode->i_rdev),boardnr=minor-1,result;
 if(minor==0||_IOC_TYPE(cmd)!=MDSP_MAJOR) return -EINVAL;
 if(minor>0&&minor<=MDSP_MAX&&mdsp_board[boardnr].ready)
  {
   __u32 val,oldval;
   __u32 status;
   unsigned int address=0;
   unsigned int type=mdsp_board[boardnr].basetype;
   int pcicommand=0;
   switch(_IOC_NR(cmd))
    {
     case 'c': address=mdsp_board[boardnr].hctr; break;
     case 's': address=mdsp_board[boardnr].hstr; break;
     case 'r': status=mdsp_read_reg(mdsp_board[boardnr].hstr,type);
               if(!(status&HSTR_HRRQ)) return -EAGAIN;
               address=mdsp_board[boardnr].hrxr;
               break;
     case 't': status=mdsp_read_reg(mdsp_board[boardnr].hstr,type);
               if(!(status&HSTR_HTRQ)) return -EAGAIN;
               address=mdsp_board[boardnr].htxr;
               break;
     case 'i': address=mdsp_board[boardnr].hcvr; break;
     case 'p': pcicommand=1; break;
     default:  return -EINVAL;
    }
   if(_IOC_DIR(cmd)&_IOC_READ)
    {
     if((result=kernel_verify_area(VERIFY_WRITE, (unsigned long *)param,
                                   sizeof(unsigned long)))!=0) return result;
     if(pcicommand) /* Read PCI CCMR/CSTR */
      {
       int result;
       if((result=kernel_pcibios_read_config_dword(mdsp_board[boardnr].busnr,
                                                  mdsp_board[boardnr].devnr,
               /* dword -> COMMAND and STATUS */  PCI_COMMAND,
                                                  &oldval))) return result;
      }
     else /* Read MDSP register */
       oldval=mdsp_read_reg(address,type);
    }
   if(_IOC_DIR(cmd)&_IOC_WRITE)
    {
     if((result=kernel_verify_area(VERIFY_READ, (unsigned long *)param,
                                   sizeof(unsigned long)))!=0) return result;
     val=(__u32)kernel_get_user((unsigned long *)param);
     if(pcicommand) /* Read PCI CCMR/CSTR */
      {
       int result;
       if((result=kernel_pcibios_write_config_dword(mdsp_board[boardnr].busnr,
                                                   mdsp_board[boardnr].devnr,
                /* dword -> COMMAND and STATUS */  PCI_COMMAND,
                                                   val))) return result;
      }
     else /* Write MDSP register */
       mdsp_write_reg(address,val,type);
    }
   if(_IOC_DIR(cmd)&_IOC_READ) /* write remembered value */
     kernel_put_user((unsigned long)oldval,(unsigned long *)param);
   return 0;
  } /* end if minor>0&&minor<=MDSP_MAX) */
 return -ENODEV;
}

/* Per default, Linux returns a successful lseek if set to NULL in          */
/* file_operations. The purpose of this function is to return a correct     */
/* error message to the caller.                                             */
static int lseek_mdsp(struct inode *inode, struct file *file, off_t offset,
                      int origin)
{
 return -ESPIPE;
}

static struct file_operations fops_mdsp =
{
 lseek_mdsp,              /* dummy lseek()             */
 read_mdsp,
 write_mdsp,
 NULL,                    /* readdir() not implemented */
 select_mdsp,
 ioctl_mdsp,
 NULL,                    /* mmap() not implemented    */
 open_mdsp,
 release_mdsp,
 NULL,                    /* fsync() not implemented   */
 NULL,                    /* fasync() not implemented  */
 NULL,                    /* check_media_change() not implemented */
 NULL                     /* revalidate() not implemented */
};

/* Board initialization */
static int init_mdsp()
{
 int i,result;
 unsigned int base;
 
 if(!kernel_pcibios_present())
  {
   PRINTK("No PCI-BIOS present, exiting.\n");
   return -1;
  }
 for(i=0;i<MDSP_MAX;i++)
  {
   mdsp_board[i].ready=0;   /* Flag non-initialized    */
   mdsp_board[i].wait=NULL; /* Zero waitqueue          */
  }
 for(i=0;i<MDSP_MAX;i++)
  {
   int j,found=0;
   /* Find all supported devices */
   for(j=0;j<sizeof(mdsp_type)/sizeof(unsigned short);j++)
    {
     found=(!kernel_pcibios_find_device(PCI_VENDOR_ID_MOTOROLA,mdsp_type[j],
                                        i,&mdsp_board[i].busnr,
                                        &mdsp_board[i].devnr));
     if(found)
      {
       mdsp_board[i].boardtype=mdsp_type[j];
       break;
      }
    }
   if(!found) break;
   if((result=kernel_pcibios_read_config_word(mdsp_board[i].busnr,
                                              mdsp_board[i].devnr,
                                              PCI_CLASS_DEVICE,
                                              &mdsp_board[i].class)))
    {
     PRINTK("Failed to read class type from PCI board, errno=%d\n",result);
     continue;
    }
   if((result=kernel_pcibios_read_config_dword(mdsp_board[i].busnr,
                                               mdsp_board[i].devnr,
                                               PCI_BASE_ADDRESS_0,
                                               &base)))
    {
     PRINTK("Failed to read base from PCI board. errno=%d\n",result);
     continue;
    }
   if((result=kernel_pcibios_read_config_byte(mdsp_board[i].busnr,
                                              mdsp_board[i].devnr,
                                              PCI_INTERRUPT_LINE,
                                              &mdsp_board[i].irq)))
    {
     PRINTK("Failed to read interrupt line from PCI board, errno=%d\n",result);
     continue;
    }
   /* Enable Memory space access and parity error handling */
   kernel_pcibios_write_config_word(mdsp_board[i].busnr,
                                     mdsp_board[i].devnr,
                                     PCI_COMMAND,
                                     PCI_COMMAND_MEMORY|
                                     PCI_COMMAND_PARITY);
   /* Memory-mapping if necessary, otherwise register IO-space */
   if(base&PCI_BASE_ADDRESS_SPACE) /* Normal IO-Port */
    {
     DPRINTK("ioaddress is already io-mapped.\n");
     if(kernel_check_region(base,MDSP_MEM_SIZE))
      {
       PRINTK("Cannot access IO memory at 0x%x, skipping board.\n",
              (unsigned int)base);
       continue;
      }
     /* Register IO region  to /proc/ioports */
     base&=PCI_BASE_ADDRESS_IO_MASK;
     kernel_request_region(base,MDSP_MEM_SIZE,mdsp_name);
     mdsp_board[i].base=base;
     mdsp_board[i].basetype=PCI_BASE_ADDRESS_SPACE_IO;
    }
   else /* Memory-mapped, must be mapped into Kernel-Space */
    {
     DPRINTK("ioaddress is memory-mapped, mapping into kernel-space.\n");
     base &= PCI_BASE_ADDRESS_MEM_MASK;
     mdsp_board[i].base = (unsigned int) kernel_ioremap(base,MDSP_MEM_SIZE);
     mdsp_board[i].basetype=PCI_BASE_ADDRESS_SPACE_MEMORY;
    }
   /* Precalculate addresses for easier access in functions. */
   mdsp_board[i].hctr=mdsp_board[i].base+HCTR;
   mdsp_board[i].hstr=mdsp_board[i].base+HSTR;
   mdsp_board[i].hcvr=mdsp_board[i].base+HCVR;
   mdsp_board[i].htxr=mdsp_board[i].base+HTXR;
   mdsp_board[i].hrxr=mdsp_board[i].base+HRXR;
   mdsp_board[i].ready=1; /* Everything ok, flag board ready */
   /* Print status of card to console */
   PRINTK("BUS: %d, DEVICE: 0x%x, CLASS: 0x%x, ADDRESS: 0x%x, IRQ: %d\n",
          (int)mdsp_board[i].busnr,(int)mdsp_board[i].devnr,
          (int)mdsp_board[i].class, (int)mdsp_board[i].base,
          (int)mdsp_board[i].irq);
  } /* for i < MDSP_MAX */
 PRINTK("Found %d Motorola DSP56xxx board%s.\n",i,i!=1?"s":"");
 return 0;
}

/* Module initialization at loading time of driver */
int init_module(void)
{
 int result;
 if(init_mdsp())
  {
   PRINTK("No DSP boards found.\n");
   return -EIO;
  }
 /* Register file operations for mdsp device */
 if((result=kernel_register_chrdev(mdsp_major, mdsp_name, &fops_mdsp))==-EBUSY)
  {
   PRINTK("Unable to get major %d for MDSP device, errno=%d.\n",mdsp_major,
          result);
   return -EIO;
  }
 PRINTK("Registered Device %s, major %d.\n", mdsp_name, mdsp_major);
 return 0;
}

/* Cleanup before unloading module, release io or unmap memory space */
static void cleanup_mdsp(void)
{
 int i;
 for(i=0;i<MDSP_MAX;i++)
  {
   if(mdsp_board[i].ready)
    {
     if(mdsp_board[i].basetype==PCI_BASE_ADDRESS_SPACE_IO)
       kernel_release_region((unsigned int)mdsp_board[i].base,MDSP_MEM_SIZE);
     else
       kernel_iounmap((void *)mdsp_board[i].base);
    }
   mdsp_board[i].ready=0;
  }
}

void cleanup_module(void)
{
 kernel_unregister_chrdev(mdsp_major,mdsp_name);
 cleanup_mdsp();
 PRINTK("Unregistered Device %s, major %d.\n", mdsp_name, mdsp_major);
}
