/**************************************************************************\
 * Example program for use with the MDSP driver. Can be used for          *
 * uploading a DSP-program in PCI bootstrap mode or for reading/writing   *
 * of registers of the DSP56301 board.                                    *
 * Copyright (C) by Klaus Knopper, March 1998                             *
\**************************************************************************/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/pci.h>
#include "mdsp.h" /* Macros, addresses, features */

#define VERSION "MDSPTOOL (C) Klaus Knopper, March 1998"

int usage(int argc, char **argv, char *msg)
{
 const char *help=
 VERSION"\n"
 "Usage: %s options\nList of options:\n"
 "  -#            Apply following command to DSP-Card number # (default 1)\n"
 "  -a            Print information for all MDSP boards from /dev/mdspstat\n"
 "  -c            Print contents of host control register\n"
 "  -C 0xnnnnnnnn Write hex number nnnnnnnn to host control register\n"
 "  -d            Dump data from DSP (forever loop)\n"
 "  -D 0xnnnnnnnn Read 0xn data words from DSP\n"
 "  -f file       Upload hexdump file in bootstrap mode (first thing to do\n"
 "                after a reset, see documentation for file format.\n"
 "                Use - for stdin and make sure that HCTR_HF0 is cleared.)\n"
 "  -i            Print contents of host vector control register\n"
 "  -I 0xnnnnnnnn Write hex number nnnnnnnn to host vector control register\n"
 "  -p            Print contents of pci control register\n"
 "  -P 0xnnnnnnnn Write hex number nnnnnnnn to pci control register\n"
 "  -t            Print contents of data transfer register immediate\n"
 "  -T 0xnnnnnnnn Write hex number nnnnnnnn to data transfer register "
 "immediate\n"
 "  -s            Print contents of host statusregister\n"
 "  -S 0xnnnnnnnn Write hex number nnnnnnnn to host status register\n"
 "  -v            Be verbose (print decoded values of registers)\n";
 fprintf(stderr,help,argv[0]);
 if(msg)
  {
   fprintf(stderr,"\nBad option: %s\n",msg);
   return -1;
  }
 return 0;
}

/* Opens MDSP-device with the given ID n, returns result of open() */
int mreopen(int fd,int n,int verbose)
{
 const char *common="/dev/mdsp";
 char mname[32]="";
 if(fd>=0) close(fd);
 sprintf(mname,"%s%d",common,n);
 if((fd=open(mname,O_RDWR))<0)
   fprintf(stderr,"\nERROR: cannot open MDSP device '%s': %s\n",
           mname,sys_errlist[errno]);
 return(fd);
}

unsigned long parse_hex(char *string,__u32 *valptr)
{
 int result;
 unsigned long val;
 if((result=sscanf(string,"0x%lx",&val))!=1)
  {
   fprintf(stderr,"\nERROR: %s is not a hex number (like '0xfefefefe')\n",
           string);
  }
 *valptr=(__u32) val;
 return result;
}

int printreg(int fd,unsigned int cmd,int verbose)
{
 unsigned long val;
 if(ioctl(fd,cmd,&val))
  {
   perror("Can't ioctl()");
   return -1;
  }
 /* print it. */
 printf("0x%lx",val);
 if(verbose)
  {
   switch(cmd)
    {
     case MDSPCTL_READ_HSTR:
       if(val&HSTR_TRDY) printf(" HSTR_TRDY");
       if(val&HSTR_HTRQ) printf(" HSTR_HTRQ");
       if(val&HSTR_HRRQ) printf(" HSTR_HRRQ");
       if(val&HSTR_HF3)  printf(" HSTR_HF3");
       if(val&HSTR_HF4)  printf(" HSTR_HF4");
       if(val&HSTR_HF5)  printf(" HSTR_HF5");
       if(val&HSTR_HINT) printf(" HSTR_HINT");
       if(val&HSTR_HREQ) printf(" HSTR_HREQ");
       break;
     case MDSPCTL_READ_HCTR:
       if(val&HCTR_TREQ) printf(" HCTR_TREQ");
       if(val&HCTR_RREQ) printf(" HCTR_RREQ");
       if(val&HCTR_HF0)  printf(" HCTR_HF0");
       if(val&HCTR_HF1)  printf(" HCTR_HF1");
       if(val&HCTR_HF2)  printf(" HCTR_HF2");
       if(val&HCTR_DMAE) printf(" HCTR_DMAE");
       if(val&HCTR_SFT)  printf(" HCTR_SFT");
       if(val&HCTR_HTF0) printf(" HCTR_HTF0");
       if(val&HCTR_HTF1) printf(" HCTR_HTF1");
       if(val&HCTR_HRF0) printf(" HCTR_HRF0");
       if(val&HCTR_HRF1) printf(" HCTR_HRF1");
       if(val&HCTR_HS0)  printf(" HCTR_HS0");
       if(val&HCTR_HS1)  printf(" HCTR_HS1");
       if(val&HCTR_HS2)  printf(" HCTR_HS2");
       if(val&HCTR_TWSD) printf(" HCTR_TWSD");
       break;
    }
  }
 printf("\n");
 return 0;
}

int writereg(int fd,unsigned long val,unsigned int cmd,int verbose)
{
 if(ioctl(fd,cmd,&val))
  {
   perror("Can't ioctl()");
   return -1;
  }
 return 0;
} 

/* mupload() reads Hexdump, writes 32-bit values to the transfer register  */
/* of the DSP board. The byte order is automatically corrected to match    */
/* the byte order of the DSP when accessing PCI registers.                 */
int mupload(int fd,char *fname,int verbose)
{
 FILE *in;
 unsigned long val;
 unsigned long control;
 if(fname==NULL||fd<0)
  {
   fprintf(stderr,"\nInvalid argument to upload(): FD=%d, fname='%s'\n",
           fd,fname?fname:"(null)");
   return -1;
  }
 in=(fname[0]=='-'&&fname[1]==0)?stdin:fopen(fname,"r");
 if(in==NULL)
  {
   fprintf(stderr,"\nERROR: Can't open file '%s' for upload.\n",
           fname);
   return -1;
  }
 /* Set Host Transfer Mode 0 */
 /* Read control register */
 if(ioctl(fd,MDSPCTL_READ_HCTR,&control))
  {
   perror("Can't ioctl()");
   if(in!=stdin) fclose(in);
   return -1;
  }
 /* Clear HTF1, set HTF0 -> 3 Byte in-a-block transfer   */
 /* Clear HF0! (HF0 signals end of upload)               */
 control=((control&~HCTR_HTF1)&~HCTR_HF0)|HCTR_HTF0;
 ioctl(fd,MDSPCTL_WRITE_HCTR,&control);
 /* I wish we could reset the DSP before doing this. :(  */
 while(!feof(in)&&!ferror(in)&&fscanf(in,"%lx",&val)==1)
  {
   if(write(fd,&val,4)!=4)
    {
     fprintf(stderr,"\nERROR: Can't write to DSP.\n");
     if(in!=stdin) fclose(in);
     return -1;
    }
  }
 if(in!=stdin) fclose(in);
 /* Force start of program (set HF0, Bootstrap loader will finish download */
 /* and start program immediately)                                         */
 control=(control&(~HCTR_HRF1))|HCTR_HRF0|HCTR_HF0;
 ioctl(fd,MDSPCTL_WRITE_HCTR,&control);
 return 0;
}

int printstatus(void)
{
 int fd;
 const char *sname="/dev/mdspstat";
 char buffer[4096];
 int bytes;
 if((fd=open(sname,O_RDONLY))<=0) return -1;
 if((bytes=read(fd,buffer,4096))<=0)
  {
   close(fd);
   return -1;
  }
 close(fd);
 fwrite(buffer,bytes,1,stdout);
 return 0;
}

int main(int argc,char **argv)
{
 int mnum=1,mfd=-1,    /* DSP Device # and file descriptor */
     verbose=0,result,i;
 
 if(argc<2||argv[1][1]=='h') return(usage(argc,argv,NULL));
 for(i=1;i<argc;i++)
  {
   __u32 val; int max=0,j;
   if(argv[i][0]!='-') return(usage(argc,argv,argv[i]));
   switch(argv[i][1])
    {
     case  0 : break;
     case 'a': if(printstatus())
                {
                 fprintf(stderr,"Can't read /dev/mdspstat.\n");
                 return errno;
                }
               break;
     case 'v': verbose=1; break;
     case 'c': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               if(printreg(mfd,MDSPCTL_READ_HCTR,verbose)) return errno;
               break;
     case 'C': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               ++i;
               if(i==argc||argv[i]==NULL||parse_hex(argv[i],&val)!=1)
                 return(usage(argc,argv,argv[i]));
               if(writereg(mfd,val,MDSPCTL_WRITE_HCTR,verbose)) return errno;
               break;
     case 'D': ++i;
               if(i==argc||argv[i]==NULL||parse_hex(argv[i],&val)!=1)
                 return(usage(argc,argv,argv[i]));
               max=(int)val; /* NO break here. */
     case 'd': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               for(j=0;(max==0)||j<max;j++)
                {
                 __u32 data;
                 if(read(mfd,&data,sizeof(__u32))!=sizeof(__u32)) return errno;
                 printf("0x%x\n",(unsigned int)data);
                }
               break;
     case 'f': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               if(i==argc) return(usage(argc,argv,argv[i]));
               if((result=mupload(mfd,argv[++i],verbose))) return result;
               break;
     case 'i': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               if(printreg(mfd,MDSPCTL_READ_HCVR,verbose)) return errno;
               break;
     case 'I': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               ++i;
               if(i==argc||argv[i]==NULL||parse_hex(argv[i],&val)!=1)
                 return(usage(argc,argv,argv[i]));
               if(writereg(mfd,val,MDSPCTL_WRITE_HCVR,verbose)) return errno;
               break;
     case 's': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               if(printreg(mfd,MDSPCTL_READ_HSTR,verbose)) return errno;
               break;
     case 'S': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               ++i;
               if(i==argc||argv[i]==NULL||parse_hex(argv[i],&val)!=1)
                 return(usage(argc,argv,argv[i]));
               if(writereg(mfd,val,MDSPCTL_WRITE_HSTR,verbose)) return errno;
               break;
     case 't': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               if(printreg(mfd,MDSPCTL_READ_HRXR,verbose)) return errno;
               break;
     case 'T': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               ++i;
               if(i==argc||argv[i]==NULL||parse_hex(argv[i],&val)!=1)
                 return(usage(argc,argv,argv[i]));
               if(writereg(mfd,val,MDSPCTL_WRITE_HTXR,verbose)) return errno;
               break;
     case 'p': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               if(ioctl(mfd,MDSPCTL_READ_CCMR,&val)) return errno;
               printf("0x%lx ",(unsigned long)val);
               if(verbose) /* print decoded values, see linux/pci.h */
                {
                 if(val&PCI_COMMAND_IO)
                   printf(" PCI_COMMAND_IO");
                 if(val&PCI_COMMAND_MEMORY)
                   printf(" PCI_COMMAND_MEMORY");
                 if(val&PCI_COMMAND_MASTER)
                   printf(" PCI_COMMAND_MASTER");
                 if(val&PCI_COMMAND_SPECIAL)
                   printf(" PCI_COMMAND_SPECIAL");
                 if(val&PCI_COMMAND_INVALIDATE)
                   printf(" PCI_COMMAND_INVALIDATE");
                 if(val&PCI_COMMAND_VGA_PALETTE)
                   printf(" PCI_COMMAND_VGA_PALETTE");
                 if(val&PCI_COMMAND_PARITY)
                   printf(" PCI_COMMAND_PARITY");
                 if(val&PCI_COMMAND_WAIT)
                   printf(" PCI_COMMAND_WAIT");
                 if(val&PCI_COMMAND_SERR)
                   printf(" PCI_COMMAND_SERR");
                 if(val&PCI_COMMAND_FAST_BACK)
                   printf(" PCI_COMMAND_FAST_BACK");
                 if(val&PCI_STATUS_66MHZ<<16L)
                   printf(" PCI_STATUS_66MHZ");
                 if(val&PCI_STATUS_UDF<<16L)
                   printf(" PCI_STATUS_UDF");
                 if(val&PCI_STATUS_FAST_BACK<<16L)
                   printf(" PCI_STATUS_FAST_BACK");
                 if(val&PCI_STATUS_PARITY<<16L)
                   printf(" PCI_STATUS_PARITY");
                 if(val&PCI_STATUS_DEVSEL_FAST<<16L)
                   printf(" PCI_STATUS_DEVSEL_FAST");
                 if(val&PCI_STATUS_DEVSEL_MEDIUM<<16L)
                   printf(" PCI_STATUS_DEVSEL_MEDIUM");
                 if(val&PCI_STATUS_DEVSEL_SLOW<<16L)
                   printf(" PCI_STATUS_DEVSEL_SLOW");
                 if(val&PCI_STATUS_SIG_TARGET_ABORT<<16L)
                   printf(" PCI_STATUS_SIG_TARGET_ABORT");
                 if(val&PCI_STATUS_REC_TARGET_ABORT<<16L)
                   printf(" PCI_STATUS_REC_TARGET_ABORT");
                 if(val&PCI_STATUS_REC_MASTER_ABORT<<16L)
                   printf(" PCI_STATUS_REC_MASTER_ABORT");
                 if(val&PCI_STATUS_SIG_SYSTEM_ERROR<<16L)
                  printf(" PCI_STATUS_SIG_SYSTEM_ERROR");
                 if(val&PCI_STATUS_DETECTED_PARITY<<16L)
                   printf(" PCI_STATUS_DETECTED_PARITY");
                } /* end if verbose */
               printf("\n");
               break;
     case 'P': if(mfd<0&&(mfd=mreopen(mfd,mnum,verbose))<0) return errno;
               ++i;
               if(i==argc||argv[i]==NULL||parse_hex(argv[i],&val)!=1)
                 return(usage(argc,argv,argv[i]));
               if(ioctl(mfd,MDSPCTL_WRITE_CCMR,&val)) return errno;
               break;
     default : if((mnum=atoi(&argv[i][1]))==0)
                 return(usage(argc,argv,argv[i]));
               if((mfd=mreopen(mfd,mnum,verbose))<0) return errno;
    }
  }
 if(mfd>=0) close(mfd);
 return 0;
}
